@@ -399,7 +399,6 @@ class Hooks:
         is_constructor = fn.get("constructor")
 
         for i, p in enumerate(fn["parameters"]):
-
             if is_constructor and p["reference"] == 1:
                 x_keepalives.append((1, i + 2))
 
@@ -665,7 +664,6 @@ class Hooks:
         self._function_hook(fn, data)
 
     def class_hook(self, cls, data):
-
         if cls["parent"] is not None and cls["access_in_parent"] == "private":
             cls["data"] = ClassData(ignore=True)
             return
@@ -691,16 +689,6 @@ class Hooks:
 
         self._add_subpackage(cls, class_data)
 
-        # fix enum paths
-        for e in cls["enums"]["public"]:
-            e["x_namespace"] = e["namespace"] + "::" + cls_name + "::"
-            enum_data = self.gendata.get_cls_enum_data(
-                e.get("name"), cls_key, class_data
-            )
-            e["data"] = enum_data
-
-            self._enum_hook(e, enum_data)
-
         # update inheritance
 
         pybase_params = set()
@@ -827,6 +815,16 @@ class Hooks:
 
         cls["x_qualname"] = cls_qualname
 
+        # fix enum paths
+        for e in cls["enums"]["public"]:
+            e["x_namespace"] = cls_qualname + "::"
+            enum_data = self.gendata.get_cls_enum_data(
+                e.get("name"), cls_key, class_data
+            )
+            e["data"] = enum_data
+
+            self._enum_hook(e, enum_data)
+
         has_constructor = False
         is_polymorphic = class_data.is_polymorphic
         vcheck_fns = []
@@ -837,7 +835,6 @@ class Hooks:
             is_polymorphic = True
 
         for access in ("public", "protected", "private"):
-
             for fn in cls["methods"][access]:
                 if fn["constructor"]:
                     has_constructor = True
