@@ -39,7 +39,6 @@ class ParamData(Model):
 
 
 class BufferType(str, enum.Enum):
-
     #: The buffer must indicate that it is readable (such as bytes, or bytearray)
     IN = "in"
 
@@ -51,7 +50,6 @@ class BufferType(str, enum.Enum):
 
 
 class BufferData(Model):
-
     #: Indicates what type of python buffer is required
     type: BufferType
 
@@ -109,6 +107,13 @@ class FunctionData(Model):
     #: If True, don't wrap this, but provide a pure virtual implementation
     ignore_pure: bool = False
 
+    #: Most of the time, you will want to specify ``ignore`` instead.
+    #:
+    #: If True, don't expose this function to python. If a trampoline is supposed
+    #: to be generated, it will still be generated. You will likely want to use
+    #: ``trampoline_cpp_code`` if you specify this.
+    ignore_py: bool = False
+
     #: Generate this in an `#ifdef`
     ifdef: Optional[str] = None
     #: Generate this in an `#ifndef`
@@ -156,6 +161,9 @@ class FunctionData(Model):
     #: list is the template parameters for that function
     template_impls: Optional[List[List[str]]] = None
 
+    #: Specify custom C++ code for the virtual function trampoline
+    trampoline_cpp_code: Optional[str] = None
+
     #: Specify a transformation lambda to be used when this virtual function
     #: is called from C++. This inline code should be a lambda that has the same
     #: arguments as the original C++ virtual function, except the first argument
@@ -187,13 +195,20 @@ class FunctionData(Model):
                 value[k] = FunctionData()
         return value
 
+    @validator("virtual_xform")
+    def validate_virtual_xform(cls, v, values):
+        if v and values.get("trampoline_cpp_code"):
+            raise ValueError(
+                f"cannot specify trampoline_cpp_code and virtual_xform for the same method"
+            )
+        return v
+
 
 if not _generating_documentation:
     FunctionData.update_forward_refs()
 
 
 class PropAccess(enum.Enum):
-
     #: Determine read/read-write automatically:
     #:
     #: * If a struct/union, default to readwrite
@@ -211,7 +226,6 @@ class PropAccess(enum.Enum):
 
 
 class PropData(Model):
-
     #: If set to True, this property is not made available to python
     ignore: bool = False
 
@@ -229,7 +243,6 @@ class PropData(Model):
 
 
 class EnumValue(Model):
-
     #: If set to True, this property is not made available to python
     ignore: bool = False
 
@@ -244,7 +257,6 @@ class EnumValue(Model):
 
 
 class EnumData(Model):
-
     #: Set your own docstring for the enum
     doc: Optional[str] = None
 
@@ -265,9 +277,18 @@ class EnumData(Model):
 
     values: Dict[str, EnumValue] = {}
 
+    #: This will insert code right before the semicolon ending the enum py
+    #: definition. You can use this to easily insert additional custom values
+    #: without using the global inline_code mechanism.
+    inline_code: Optional[str] = None
+
+    #: Tell pybind11 to create an enumeration that also supports rudimentary
+    #: arithmetic and bit-level operations like comparisons, and, or, xor,
+    #: negation, etc.
+    arithmetic: bool = False
+
 
 class ClassData(Model):
-
     #: Docstring for the class
     doc: Optional[str] = None
 
@@ -277,7 +298,8 @@ class ClassData(Model):
     ignore: bool = False
     ignored_bases: List[str] = []
 
-    #: Specify fully qualified names for the bases
+    #: Specify fully qualified names for the bases. If the base has a template
+    #: parameter, you must include it
     base_qualnames: Dict[str, str] = {}
 
     attributes: Dict[str, PropData] = {}
@@ -343,6 +365,19 @@ class ClassData(Model):
     #: the bottom of the trampoline class. This is rarely useful.
     trampoline_inline_code: Optional[str] = None
 
+    #: This will insert code right before the semicolon ending the class py
+    #: definition. You can use this to easily insert additional custom functions
+    #: without using the global inline_code mechanism.
+    #:
+    #: .. code-block: C++
+    #:
+    #:   cls_MyClass
+    #:      .def("something", &MyClass::Something)  // autogenerated
+    #:      // inline code is inserted here
+    #:      ;
+    #:
+    inline_code: Optional[str] = None
+
     @validator("attributes", pre=True)
     def validate_attributes(cls, value):
         for k, v in value.items():
@@ -500,6 +535,10 @@ class HooksDataYaml(Model):
     #:
     templates: Dict[str, TemplateData] = {}
 
+    #: Extra 'using' directives to insert into the trampoline and the
+    #: wrapping scope
+    typealias: List[str] = []
+
     @validator("attributes", pre=True)
     def validate_attributes(cls, value):
         for k, v in value.items():
