@@ -17,7 +17,7 @@ from os.path import (
 import posixpath
 import shutil
 import toposort
-from typing import Any, Dict, List, Optional, Set
+from typing import Any, Dict, List, Optional, Set, Tuple
 import yaml
 
 from header2whatever.config import Config
@@ -46,7 +46,6 @@ class Wrapper:
     _cpp_version = "__cplusplus 201703L"
 
     def __init__(self, package_name, cfg: WrapperConfig, setup):
-
         self.package_name = package_name
         self.cfg = cfg
 
@@ -84,7 +83,7 @@ class Wrapper:
         self.depends = self.cfg.depends
 
         # Files that are generated AND need to be in the final wheel. Used by build_py
-        self.generated_files: List[str] = []
+        self.additional_data_files: List[str] = []
 
         self._all_deps = None
 
@@ -131,6 +130,8 @@ class Wrapper:
 
         self.dev_config = get_dev_config(self.name)
 
+        self._update_addl_data_files()
+
     def _extract_zip_to(self, dl: Download, dst, cache):
         try:
             download_and_extract_zip(dl.url, dst, cache)
@@ -174,9 +175,9 @@ class Wrapper:
                     raise OSError(err_msg)
                 raise e
 
-    def _add_generated_file(self, fullpath):
+    def _add_addl_data_file(self, fullpath):
         if not isdir(fullpath):
-            self.generated_files.append(relpath(fullpath, self.root))
+            self.additional_data_files.append(relpath(fullpath, self.root))
 
     # pkgcfg interface
     def get_include_dirs(self) -> Optional[List[str]]:
@@ -248,6 +249,20 @@ class Wrapper:
             for typ in ccfg.types:
                 casters[typ] = cfg
 
+    def _update_addl_data_files(self) -> List[str]:
+        headers = set()
+        for ccfg in self.cfg.type_casters:
+            headers.add(ccfg.header)
+
+        if headers:
+            includes = self.get_include_dirs()
+            if includes:
+                for hdr in headers:
+                    for p in includes:
+                        fpath = join(p, hdr)
+                        if exists(fpath):
+                            self._add_addl_data_file(fpath)
+
     def all_deps(self):
         if self._all_deps is None:
             self._all_deps = self.pkgcfg.get_all_deps(self.name)
@@ -261,6 +276,9 @@ class Wrapper:
             includes.extend(self.pkgcfg.get_pkg("robotpy-build").get_include_dirs())
         return includes
 
+    def _generation_search_path(self):
+        return [self.root] + self._all_includes(False)
+
     def _all_library_dirs(self):
         libs = self.get_library_dirs()
         for dep in self.cfg.depends:
@@ -311,7 +329,6 @@ class Wrapper:
         return casters
 
     def on_build_dl(self, cache: str, srcdir: str):
-
         pkgcfgpy = join(self.root, "pkgcfg.py")
         srcdir = join(srcdir, self.name)
 
@@ -326,13 +343,16 @@ class Wrapper:
             pass
 
         libnames_full = []
+        all_libs = []
         downloads = self.cfg.download
         if downloads:
-            libnames_full = self._clean_and_download(downloads, cache, srcdir)
+            libnames_full, all_libs = self._clean_and_download(downloads, cache, srcdir)
 
         self._write_libinit_py(libnames_full)
         self._write_pkgcfg_py(pkgcfgpy, libnames_full)
 
+        return all_libs
+
     def _apply_patches(self, patches: List[PatchInfo], root: str):
         import patch
 
@@ -348,8 +368,7 @@ class Wrapper:
 
     def _clean_and_download(
         self, downloads: List[Download], cache: str, srcdir: str
-    ) -> List[str]:
-
+    ) -> Tuple[List[str], List[str]]:
         libdir = join(self.root, "lib")
         incdir = join(self.root, "include")
 
@@ -363,6 +382,7 @@ class Wrapper:
 
         dlopen_libnames = self.get_dlopen_library_names()
         libnames_full = []
+        all_libs = []
 
         for dl in downloads:
             # extract the whole thing into a directory when using for sources
@@ -407,6 +427,7 @@ class Wrapper:
                     posixpath.join(dl.libdir, libname): join(libdir, libname)
                     for libname in extract_names
                 }
+                all_libs.extend(to.values())
             else:
                 to = {}
 
@@ -421,16 +442,15 @@ class Wrapper:
 
         if add_incdir:
             for f in glob.glob(join(glob.escape(incdir), "**"), recursive=True):
-                self._add_generated_file(f)
+                self._add_addl_data_file(f)
 
         if add_libdir:
             for f in glob.glob(join(glob.escape(libdir), "**"), recursive=True):
-                self._add_generated_file(f)
+                self._add_addl_data_file(f)
 
-        return libnames_full
+        return libnames_full, all_libs
 
     def _write_libinit_py(self, libnames):
-
         # This file exists to ensure that any shared library dependencies
         # are loaded for the compiled extension
 
@@ -481,10 +501,9 @@ class Wrapper:
         with open(self.libinit_import_py, "w") as fp:
             fp.write(init)
 
-        self._add_generated_file(self.libinit_import_py)
+        self._add_addl_data_file(self.libinit_import_py)
 
     def _write_pkgcfg_py(self, fname, libnames_full):
-
         library_dirs = "[]"
         library_dirs_rel = []
         library_names = self.get_library_names()
@@ -562,7 +581,7 @@ class Wrapper:
         with open(fname, "w") as fp:
             fp.write(pkgcfg)
 
-        self._add_generated_file(fname)
+        self._add_addl_data_file(fname)
 
     def _load_generation_data(self, datafile):
         with open(datafile) as fp:
@@ -576,7 +595,6 @@ class Wrapper:
     def on_build_gen(
         self, cxx_gen_dir, missing_reporter: Optional[MissingReporter] = None
     ):
-
         if not self.cfg.autogen_headers:
             return
 
@@ -593,6 +611,8 @@ class Wrapper:
         hppoutdir = join(self.rpy_incdir, "rpygen")
         tmpl_dir = join(thisdir, "templates")
         cpp_tmpl = join(tmpl_dir, "cls.cpp.j2")
+        cls_tmpl_inst_cpp = join(tmpl_dir, "cls_tmpl_inst.cpp.j2")
+        cls_tmpl_inst_hpp = join(tmpl_dir, "cls_tmpl_inst.hpp.j2")
         hpp_tmpl = join(tmpl_dir, "cls_rpy_include.hpp.j2")
         classdeps_tmpl = join(tmpl_dir, "clsdeps.json.j2")
 
@@ -601,7 +621,6 @@ class Wrapper:
         # TODO: only regenerate files if the generated files
         #       have changed
         if not report_only:
-
             if self.dev_config.only_generate is None:
                 shutil.rmtree(cxx_gen_dir, ignore_errors=True)
                 shutil.rmtree(hppoutdir, ignore_errors=True)
@@ -632,10 +651,9 @@ class Wrapper:
         else:
             only_generate = None
 
-        generation_search_path = [self.root] + self._all_includes(False)
+        generation_search_path = self._generation_search_path()
 
         for name, header in self.cfg.autogen_headers.items():
-
             header = normpath(header)
             for path in generation_search_path:
                 header_path = join(path, header)
@@ -667,9 +685,6 @@ class Wrapper:
                 ]
                 class_templates = [{"src": hpp_tmpl, "dst": hpp_dst}]
 
-            if only_generate is not None and not only_generate.pop(name, False):
-                continue
-
             if per_header:
                 data_fname = join(datapath, name + ".yml")
                 if not exists(data_fname):
@@ -678,6 +693,30 @@ class Wrapper:
                 else:
                     data = self._load_generation_data(data_fname)
 
+                # split instantiation of each template to separate cpp files to reduce
+                # compiler memory for really obscene objects
+                if not report_only and data.templates:
+                    class_templates.append(
+                        {
+                            "src": cls_tmpl_inst_hpp,
+                            "dst": join(cxx_gen_dir, f"{name}_tmpl.hpp"),
+                        }
+                    )
+
+                    for i, k in enumerate(data.templates.keys(), start=1):
+                        tmpl_cpp_dst = join(cxx_gen_dir, f"{name}_tmpl{i}.cpp")
+                        class_templates.append(
+                            {
+                                "src": cls_tmpl_inst_cpp,
+                                "dst": tmpl_cpp_dst,
+                                "vars": {"index": i, "key": k},
+                            }
+                        )
+                        self.extension.sources.append(tmpl_cpp_dst)
+
+            if only_generate is not None and not only_generate.pop(name, False):
+                continue
+
             # for each thing, create a h2w configuration dictionary
             cfgd = {
                 # generation code depends on this being just one header!
@@ -723,7 +762,7 @@ class Wrapper:
         self._gen_includes = gen_includes
 
         for f in glob.glob(join(glob.escape(hppoutdir), "*.hpp")):
-            self._add_generated_file(f)
+            self._add_addl_data_file(f)
 
     def finalize_extension(self):
         if self.extension is None:
@@ -743,7 +782,6 @@ class Wrapper:
         self.extension.extra_objects = self._all_extra_objects()
 
     def _write_wrapper_hpp(self, outdir, classdeps):
-
         decls = []
         begin_calls = []
         finish_calls = []
